/****************************************************************

  Generated by Eclipse Cyclone DDS IDL to C Translator
  File name: ddsi_xt_typeinfo.c
  Source: ddsi_xt_typeinfo.idl
  Cyclone DDS: V0.9.0

*****************************************************************/
#include "dds/ddsi/ddsi_xt_typeinfo.h"

static const uint32_t DDS_XTypes_TypeIdentifier_ops [] =
{
  /* TypeIdentifier */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeIdentifier, _d), 13u, (60u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 57 /* StringSTypeDefn */, 112, offsetof (DDS_XTypes_TypeIdentifier, _u.string_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 53 /* StringSTypeDefn */, 114, offsetof (DDS_XTypes_TypeIdentifier, _u.string_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 52 /* StringLTypeDefn */, 113, offsetof (DDS_XTypes_TypeIdentifier, _u.string_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 48 /* StringLTypeDefn */, 115, offsetof (DDS_XTypes_TypeIdentifier, _u.string_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 47 /* PlainSequenceSElemDefn */, 128, offsetof (DDS_XTypes_TypeIdentifier, _u.seq_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 60 /* PlainSequenceLElemDefn */, 129, offsetof (DDS_XTypes_TypeIdentifier, _u.seq_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 66 /* PlainArraySElemDefn */, 144, offsetof (DDS_XTypes_TypeIdentifier, _u.array_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 72 /* PlainArrayLElemDefn */, 145, offsetof (DDS_XTypes_TypeIdentifier, _u.array_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 78 /* PlainMapSTypeDefn */, 160, offsetof (DDS_XTypes_TypeIdentifier, _u.map_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 92 /* PlainMapLTypeDefn */, 161, offsetof (DDS_XTypes_TypeIdentifier, _u.map_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 106 /* StronglyConnectedComponentId */, 176, offsetof (DDS_XTypes_TypeIdentifier, _u.sc_component_id), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 8, 242, offsetof (DDS_XTypes_TypeIdentifier, _u.equivalence_hash), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 4, 241, offsetof (DDS_XTypes_TypeIdentifier, _u.equivalence_hash), 0u,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, 0u, 14u,
  DDS_OP_RTS,
  DDS_OP_RTS,

  /* StringSTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_StringSTypeDefn, bound),
  DDS_OP_RTS,

  /* StringLTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_StringLTypeDefn, bound),
  DDS_OP_RTS,

  /* PlainSequenceSElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceSElemDefn, header), (3u << 16u) + 10u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainSequenceSElemDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceSElemDefn, element_identifier), (4u << 16u) + 65464u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainCollectionHeader */
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainCollectionHeader, equiv_kind),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainCollectionHeader, element_flags), 0u, 127u,
  DDS_OP_RTS,

  /* PlainSequenceLElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceLElemDefn, header), (3u << 16u) + 65529u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_PlainSequenceLElemDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceLElemDefn, element_identifier), (4u << 16u) + 65447u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainArraySElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArraySElemDefn, header), (3u << 16u) + 65519u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_PlainArraySElemDefn, array_bound_seq),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArraySElemDefn, element_identifier), (4u << 16u) + 65437u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainArrayLElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArrayLElemDefn, header), (3u << 16u) + 65509u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_4BY, offsetof (DDS_XTypes_PlainArrayLElemDefn, array_bound_seq),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArrayLElemDefn, element_identifier), (4u << 16u) + 65427u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainMapSTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, header), (3u << 16u) + 65499u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainMapSTypeDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, element_identifier), (4u << 16u) + 65417u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainMapSTypeDefn, key_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, key_identifier), (4u << 16u) + 65409u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainMapLTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, header), (3u << 16u) + 65481u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_PlainMapLTypeDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, element_identifier), (4u << 16u) + 65399u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainMapLTypeDefn, key_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, key_identifier), (4u << 16u) + 65391u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* StronglyConnectedComponentId */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_StronglyConnectedComponentId, sc_component_id), (3u << 16u) + 8u /* TypeObjectHashId */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_StronglyConnectedComponentId, scc_length),
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_StronglyConnectedComponentId, scc_index),
  DDS_OP_RTS,

  /* TypeObjectHashId */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeObjectHashId, _d), 2u, (16u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 8, 242, offsetof (DDS_XTypes_TypeObjectHashId, _u.hash), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 4, 241, offsetof (DDS_XTypes_TypeObjectHashId, _u.hash), 0u,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, 0u, 14u,
  DDS_OP_RTS,
  DDS_OP_RTS
};

const dds_topic_descriptor_t DDS_XTypes_TypeIdentifier_desc =
{
  .m_size = sizeof (DDS_XTypes_TypeIdentifier),
  .m_align = dds_alignof (DDS_XTypes_TypeIdentifier),
  .m_flagset = DDS_TOPIC_CONTAINS_UNION,
  .m_nkeys = 0u,
  .m_typename = "DDS::XTypes::TypeIdentifier",
  .m_keys = NULL,
  .m_nops = 63,
  .m_ops = DDS_XTypes_TypeIdentifier_ops,
  .m_meta = ""
};

static const uint32_t DDS_XTypes_TypeObject_ops [] =
{
  /* TypeObject */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeObject, _d), 2u, (12u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_UNI | 9 /* CompleteTypeObject */, 242, offsetof (DDS_XTypes_TypeObject, _u.complete), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_UNI | 730 /* MinimalTypeObject */, 241, offsetof (DDS_XTypes_TypeObject, _u.minimal), 0u,
  DDS_OP_RTS,

  /* CompleteTypeObject */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY | DDS_OP_FLAG_DEF, offsetof (DDS_XTypes_CompleteTypeObject, _d), 11u, (48u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 45 /* CompleteAliasType */, 48, offsetof (DDS_XTypes_CompleteTypeObject, _u.alias_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 379 /* CompleteAnnotationType */, 80, offsetof (DDS_XTypes_CompleteTypeObject, _u.annotation_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 411 /* CompleteStructType */, 81, offsetof (DDS_XTypes_CompleteTypeObject, _u.struct_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 457 /* CompleteUnionType */, 82, offsetof (DDS_XTypes_CompleteTypeObject, _u.union_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 514 /* CompleteBitsetType */, 83, offsetof (DDS_XTypes_CompleteTypeObject, _u.bitset_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 547 /* CompleteSequenceType */, 96, offsetof (DDS_XTypes_CompleteTypeObject, _u.sequence_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 591 /* CompleteArrayType */, 97, offsetof (DDS_XTypes_CompleteTypeObject, _u.array_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 610 /* CompleteMapType */, 98, offsetof (DDS_XTypes_CompleteTypeObject, _u.map_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 620 /* CompleteEnumeratedType */, 64, offsetof (DDS_XTypes_CompleteTypeObject, _u.enumerated_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 655 /* CompleteBitmaskType */, 65, offsetof (DDS_XTypes_CompleteTypeObject, _u.bitmask_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 679 /* CompleteExtendedType */, 0, 0u, 0u,
  DDS_OP_RTS,

  /* CompleteAliasType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteAliasType, alias_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAliasType, header), (3u << 16u) + 7u /* CompleteAliasHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAliasType, body), (3u << 16u) + 296u /* CompleteAliasBody */,
  DDS_OP_RTS,

  /* CompleteAliasHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAliasHeader, detail), (3u << 16u) + 4u /* CompleteTypeDetail */,
  DDS_OP_RTS,

  /* CompleteTypeDetail */
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteTypeDetail, ann_builtin), (4u << 16u) + 12u /* AppliedBuiltinTypeAnnotations */, sizeof (DDS_XTypes_AppliedBuiltinTypeAnnotations),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteTypeDetail, ann_custom), sizeof (DDS_XTypes_AppliedAnnotation), (4u << 16u) + 23u /* AppliedAnnotation */,
  DDS_OP_ADR | DDS_OP_TYPE_BST, offsetof (DDS_XTypes_CompleteTypeDetail, type_name), 257u,
  DDS_OP_RTS,

  /* AppliedBuiltinTypeAnnotations */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_AppliedBuiltinTypeAnnotations, verbatim), (4u << 16u) + 5u /* AppliedVerbatimAnnotation */, sizeof (DDS_XTypes_AppliedVerbatimAnnotation),
  DDS_OP_RTS,

  /* AppliedVerbatimAnnotation */
  DDS_OP_ADR | DDS_OP_TYPE_BST, offsetof (DDS_XTypes_AppliedVerbatimAnnotation, placement), 33u,
  DDS_OP_ADR | DDS_OP_TYPE_BST, offsetof (DDS_XTypes_AppliedVerbatimAnnotation, language), 33u,
  DDS_OP_ADR | DDS_OP_TYPE_STR, offsetof (DDS_XTypes_AppliedVerbatimAnnotation, text),
  DDS_OP_RTS,

  /* AppliedAnnotation */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_AppliedAnnotation, annotation_typeid), (3u << 16u) + 8u /* TypeIdentifier */,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_AppliedAnnotation, param_seq), sizeof (DDS_XTypes_AppliedAnnotationParameter), (4u << 16u) + 181u /* AppliedAnnotationParameter */,
  DDS_OP_RTS,

  /* TypeIdentifier */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeIdentifier, _d), 13u, (60u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 57 /* StringSTypeDefn */, 112, offsetof (DDS_XTypes_TypeIdentifier, _u.string_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 53 /* StringSTypeDefn */, 114, offsetof (DDS_XTypes_TypeIdentifier, _u.string_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 52 /* StringLTypeDefn */, 113, offsetof (DDS_XTypes_TypeIdentifier, _u.string_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 48 /* StringLTypeDefn */, 115, offsetof (DDS_XTypes_TypeIdentifier, _u.string_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 47 /* PlainSequenceSElemDefn */, 128, offsetof (DDS_XTypes_TypeIdentifier, _u.seq_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 60 /* PlainSequenceLElemDefn */, 129, offsetof (DDS_XTypes_TypeIdentifier, _u.seq_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 66 /* PlainArraySElemDefn */, 144, offsetof (DDS_XTypes_TypeIdentifier, _u.array_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 72 /* PlainArrayLElemDefn */, 145, offsetof (DDS_XTypes_TypeIdentifier, _u.array_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 78 /* PlainMapSTypeDefn */, 160, offsetof (DDS_XTypes_TypeIdentifier, _u.map_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 92 /* PlainMapLTypeDefn */, 161, offsetof (DDS_XTypes_TypeIdentifier, _u.map_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 106 /* StronglyConnectedComponentId */, 176, offsetof (DDS_XTypes_TypeIdentifier, _u.sc_component_id), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 8, 242, offsetof (DDS_XTypes_TypeIdentifier, _u.equivalence_hash), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 4, 241, offsetof (DDS_XTypes_TypeIdentifier, _u.equivalence_hash), 0u,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, 0u, 14u,
  DDS_OP_RTS,
  DDS_OP_RTS,

  /* StringSTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_StringSTypeDefn, bound),
  DDS_OP_RTS,

  /* StringLTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_StringLTypeDefn, bound),
  DDS_OP_RTS,

  /* PlainSequenceSElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceSElemDefn, header), (3u << 16u) + 10u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainSequenceSElemDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceSElemDefn, element_identifier), (4u << 16u) + 65464u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainCollectionHeader */
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainCollectionHeader, equiv_kind),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainCollectionHeader, element_flags), 0u, 127u,
  DDS_OP_RTS,

  /* PlainSequenceLElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceLElemDefn, header), (3u << 16u) + 65529u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_PlainSequenceLElemDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceLElemDefn, element_identifier), (4u << 16u) + 65447u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainArraySElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArraySElemDefn, header), (3u << 16u) + 65519u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_PlainArraySElemDefn, array_bound_seq),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArraySElemDefn, element_identifier), (4u << 16u) + 65437u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainArrayLElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArrayLElemDefn, header), (3u << 16u) + 65509u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_4BY, offsetof (DDS_XTypes_PlainArrayLElemDefn, array_bound_seq),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArrayLElemDefn, element_identifier), (4u << 16u) + 65427u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainMapSTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, header), (3u << 16u) + 65499u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainMapSTypeDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, element_identifier), (4u << 16u) + 65417u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainMapSTypeDefn, key_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, key_identifier), (4u << 16u) + 65409u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainMapLTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, header), (3u << 16u) + 65481u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_PlainMapLTypeDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, element_identifier), (4u << 16u) + 65399u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainMapLTypeDefn, key_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, key_identifier), (4u << 16u) + 65391u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* StronglyConnectedComponentId */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_StronglyConnectedComponentId, sc_component_id), (3u << 16u) + 8u /* TypeObjectHashId */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_StronglyConnectedComponentId, scc_length),
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_StronglyConnectedComponentId, scc_index),
  DDS_OP_RTS,

  /* TypeObjectHashId */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeObjectHashId, _d), 2u, (16u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 8, 242, offsetof (DDS_XTypes_TypeObjectHashId, _u.hash), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 4, 241, offsetof (DDS_XTypes_TypeObjectHashId, _u.hash), 0u,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, 0u, 14u,
  DDS_OP_RTS,
  DDS_OP_RTS,

  /* AppliedAnnotationParameter */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_AppliedAnnotationParameter, paramname_hash), 4u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_AppliedAnnotationParameter, value), (3u << 16u) + 4u /* AnnotationParameterValue */,
  DDS_OP_RTS,

  /* AnnotationParameterValue */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY | DDS_OP_FLAG_DEF, offsetof (DDS_XTypes_AnnotationParameterValue, _d), 14u, (64u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_BLN, 1, offsetof (DDS_XTypes_AnnotationParameterValue, _u.boolean_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_1BY, 2, offsetof (DDS_XTypes_AnnotationParameterValue, _u.byte_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_2BY, 3, offsetof (DDS_XTypes_AnnotationParameterValue, _u.int16_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_2BY, 6, offsetof (DDS_XTypes_AnnotationParameterValue, _u.uint_16_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_4BY, 4, offsetof (DDS_XTypes_AnnotationParameterValue, _u.int32_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_4BY, 7, offsetof (DDS_XTypes_AnnotationParameterValue, _u.uint32_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_8BY, 5, offsetof (DDS_XTypes_AnnotationParameterValue, _u.int64_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_8BY, 8, offsetof (DDS_XTypes_AnnotationParameterValue, _u.uint64_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_4BY, 9, offsetof (DDS_XTypes_AnnotationParameterValue, _u.float32_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_8BY, 10, offsetof (DDS_XTypes_AnnotationParameterValue, _u.float64_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_1BY, 16, offsetof (DDS_XTypes_AnnotationParameterValue, _u.char_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_4BY, 64, offsetof (DDS_XTypes_AnnotationParameterValue, _u.enumerated_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_BST | 8, 32, offsetof (DDS_XTypes_AnnotationParameterValue, _u.string8_value), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 9 /* ExtendedAnnotationParameterValue */, 0, 0u, 0u,
  DDS_OP_ADR | DDS_OP_TYPE_BST, 0u, 129u,
  DDS_OP_RTS,
  DDS_OP_RTS,

  /* ExtendedAnnotationParameterValue */
  DDS_OP_PLC,
  DDS_OP_RTS,

  /* CompleteAliasBody */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAliasBody, common), (3u << 16u) + 12u /* CommonAliasBody */,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAliasBody, ann_builtin), (4u << 16u) + 17u /* AppliedBuiltinMemberAnnotations */, sizeof (DDS_XTypes_AppliedBuiltinMemberAnnotations),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteAliasBody, ann_custom), sizeof (DDS_XTypes_AppliedAnnotation), (4u << 16u) + 65268u /* AppliedAnnotation */,
  DDS_OP_RTS,

  /* CommonAliasBody */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonAliasBody, related_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CommonAliasBody, related_type), (3u << 16u) + 65268u /* TypeIdentifier */,
  DDS_OP_RTS,

  /* AppliedBuiltinMemberAnnotations */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_TYPE_STR, offsetof (DDS_XTypes_AppliedBuiltinMemberAnnotations, unit),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_AppliedBuiltinMemberAnnotations, min), (4u << 16u) + 65445u /* AnnotationParameterValue */, sizeof (DDS_XTypes_AnnotationParameterValue),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_AppliedBuiltinMemberAnnotations, max), (4u << 16u) + 65441u /* AnnotationParameterValue */, sizeof (DDS_XTypes_AnnotationParameterValue),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_TYPE_STR, offsetof (DDS_XTypes_AppliedBuiltinMemberAnnotations, hash_id),
  DDS_OP_RTS,

  /* CompleteAnnotationType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteAnnotationType, annotation_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAnnotationType, header), (3u << 16u) + 8u /* CompleteAnnotationHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteAnnotationType, member_seq), sizeof (DDS_XTypes_CompleteAnnotationParameter), (4u << 16u) + 10u /* CompleteAnnotationParameter */,
  DDS_OP_RTS,

  /* CompleteAnnotationHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_BST, offsetof (DDS_XTypes_CompleteAnnotationHeader, annotation_name), 257u,
  DDS_OP_RTS,

  /* CompleteAnnotationParameter */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAnnotationParameter, common), (3u << 16u) + 10u /* CommonAnnotationParameter */,
  DDS_OP_ADR | DDS_OP_TYPE_BST, offsetof (DDS_XTypes_CompleteAnnotationParameter, name), 257u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteAnnotationParameter, default_value), (3u << 16u) + 65410u /* AnnotationParameterValue */,
  DDS_OP_RTS,

  /* CommonAnnotationParameter */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonAnnotationParameter, member_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CommonAnnotationParameter, member_type_id), (3u << 16u) + 65218u /* TypeIdentifier */,
  DDS_OP_RTS,

  /* CompleteStructType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteStructType, struct_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteStructType, header), (3u << 16u) + 8u /* CompleteStructHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteStructType, member_seq), sizeof (DDS_XTypes_CompleteStructMember), (4u << 16u) + 13u /* CompleteStructMember */,
  DDS_OP_RTS,

  /* CompleteStructHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteStructHeader, base_type), (3u << 16u) + 65201u /* TypeIdentifier */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteStructHeader, detail), (3u << 16u) + 65162u /* CompleteTypeDetail */,
  DDS_OP_RTS,

  /* CompleteStructMember */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteStructMember, common), (3u << 16u) + 7u /* CommonStructMember */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteStructMember, detail), (3u << 16u) + 14u /* CompleteMemberDetail */,
  DDS_OP_RTS,

  /* CommonStructMember */
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_CommonStructMember, member_id),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonStructMember, member_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CommonStructMember, member_type_id), (3u << 16u) + 65180u /* TypeIdentifier */,
  DDS_OP_RTS,

  /* CompleteMemberDetail */
  DDS_OP_ADR | DDS_OP_TYPE_BST, offsetof (DDS_XTypes_CompleteMemberDetail, name), 257u,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteMemberDetail, ann_builtin), (4u << 16u) + 65445u /* AppliedBuiltinMemberAnnotations */, sizeof (DDS_XTypes_AppliedBuiltinMemberAnnotations),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteMemberDetail, ann_custom), sizeof (DDS_XTypes_AppliedAnnotation), (4u << 16u) + 65160u /* AppliedAnnotation */,
  DDS_OP_RTS,

  /* CompleteUnionType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteUnionType, union_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteUnionType, header), (3u << 16u) + 11u /* CompleteUnionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteUnionType, discriminator), (3u << 16u) + 13u /* CompleteDiscriminatorMember */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteUnionType, member_seq), sizeof (DDS_XTypes_CompleteUnionMember), (4u << 16u) + 31u /* CompleteUnionMember */,
  DDS_OP_RTS,

  /* CompleteUnionHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteUnionHeader, detail), (3u << 16u) + 65112u /* CompleteTypeDetail */,
  DDS_OP_RTS,

  /* CompleteDiscriminatorMember */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteDiscriminatorMember, common), (3u << 16u) + 12u /* CommonDiscriminatorMember */,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteDiscriminatorMember, ann_builtin), (4u << 16u) + 65116u /* AppliedBuiltinTypeAnnotations */, sizeof (DDS_XTypes_AppliedBuiltinTypeAnnotations),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteDiscriminatorMember, ann_custom), sizeof (DDS_XTypes_AppliedAnnotation), (4u << 16u) + 65127u /* AppliedAnnotation */,
  DDS_OP_RTS,

  /* CommonDiscriminatorMember */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonDiscriminatorMember, member_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CommonDiscriminatorMember, type_id), (3u << 16u) + 65127u /* TypeIdentifier */,
  DDS_OP_RTS,

  /* CompleteUnionMember */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteUnionMember, common), (3u << 16u) + 7u /* CommonUnionMember */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteUnionMember, detail), (3u << 16u) + 65479u /* CompleteMemberDetail */,
  DDS_OP_RTS,

  /* CommonUnionMember */
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_CommonUnionMember, member_id),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonUnionMember, member_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CommonUnionMember, type_id), (3u << 16u) + 65109u /* TypeIdentifier */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_CommonUnionMember, label_seq),
  DDS_OP_RTS,

  /* CompleteBitsetType */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteBitsetType, bitset_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitsetType, header), (3u << 16u) + 8u /* CompleteBitsetHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteBitsetType, field_seq), sizeof (DDS_XTypes_CompleteBitfield), (4u << 16u) + 10u /* CompleteBitfield */,
  DDS_OP_RTS,

  /* CompleteBitsetHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitsetHeader, detail), (3u << 16u) + 65053u /* CompleteTypeDetail */,
  DDS_OP_RTS,

  /* CompleteBitfield */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitfield, common), (3u << 16u) + 7u /* CommonBitfield */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitfield, detail), (3u << 16u) + 65441u /* CompleteMemberDetail */,
  DDS_OP_RTS,

  /* CommonBitfield */
  DDS_OP_ADR | DDS_OP_TYPE_2BY, offsetof (DDS_XTypes_CommonBitfield, position),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonBitfield, flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_CommonBitfield, bitcount),
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_CommonBitfield, holder_type),
  DDS_OP_RTS,

  /* CompleteSequenceType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteSequenceType, collection_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteSequenceType, header), (3u << 16u) + 7u /* CompleteCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteSequenceType, element), (3u << 16u) + 16u /* CompleteCollectionElement */,
  DDS_OP_RTS,

  /* CompleteCollectionHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteCollectionHeader, common), (3u << 16u) + 8u /* CommonCollectionHeader */,
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteCollectionHeader, detail), (4u << 16u) + 65015u /* CompleteTypeDetail */, sizeof (DDS_XTypes_CompleteTypeDetail),
  DDS_OP_RTS,

  /* CommonCollectionHeader */
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_CommonCollectionHeader, bound),
  DDS_OP_RTS,

  /* CompleteCollectionElement */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteCollectionElement, common), (3u << 16u) + 7u /* CommonCollectionElement */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteCollectionElement, detail), (3u << 16u) + 12u /* CompleteElementDetail */,
  DDS_OP_RTS,

  /* CommonCollectionElement */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonCollectionElement, element_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CommonCollectionElement, type), (3u << 16u) + 65031u /* TypeIdentifier */,
  DDS_OP_RTS,

  /* CompleteElementDetail */
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteElementDetail, ann_builtin), (4u << 16u) + 65299u /* AppliedBuiltinMemberAnnotations */, sizeof (DDS_XTypes_AppliedBuiltinMemberAnnotations),
  DDS_OP_ADR | DDS_OP_FLAG_OPT | DDS_OP_FLAG_EXT | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteElementDetail, ann_custom), sizeof (DDS_XTypes_AppliedAnnotation), (4u << 16u) + 65014u /* AppliedAnnotation */,
  DDS_OP_RTS,

  /* CompleteArrayType */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteArrayType, collection_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteArrayType, header), (3u << 16u) + 7u /* CompleteArrayHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteArrayType, element), (3u << 16u) + 65503u /* CompleteCollectionElement */,
  DDS_OP_RTS,

  /* CompleteArrayHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteArrayHeader, common), (3u << 16u) + 7u /* CommonArrayHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteArrayHeader, detail), (3u << 16u) + 64966u /* CompleteTypeDetail */,
  DDS_OP_RTS,

  /* CommonArrayHeader */
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_4BY, offsetof (DDS_XTypes_CommonArrayHeader, bound_seq),
  DDS_OP_RTS,

  /* CompleteMapType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteMapType, collection_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteMapType, header), (3u << 16u) + 65472u /* CompleteCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteMapType, key), (3u << 16u) + 65481u /* CompleteCollectionElement */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteMapType, element), (3u << 16u) + 65478u /* CompleteCollectionElement */,
  DDS_OP_RTS,

  /* CompleteEnumeratedType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteEnumeratedType, enum_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteEnumeratedType, header), (3u << 16u) + 8u /* CompleteEnumeratedHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteEnumeratedType, literal_seq), sizeof (DDS_XTypes_CompleteEnumeratedLiteral), (4u << 16u) + 16u /* CompleteEnumeratedLiteral */,
  DDS_OP_RTS,

  /* CompleteEnumeratedHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteEnumeratedHeader, common), (3u << 16u) + 7u /* CommonEnumeratedHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteEnumeratedHeader, detail), (3u << 16u) + 64929u /* CompleteTypeDetail */,
  DDS_OP_RTS,

  /* CommonEnumeratedHeader */
  DDS_OP_ADR | DDS_OP_TYPE_2BY, offsetof (DDS_XTypes_CommonEnumeratedHeader, bit_bound),
  DDS_OP_RTS,

  /* CompleteEnumeratedLiteral */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteEnumeratedLiteral, common), (3u << 16u) + 7u /* CommonEnumeratedLiteral */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteEnumeratedLiteral, detail), (3u << 16u) + 65314u /* CompleteMemberDetail */,
  DDS_OP_RTS,

  /* CommonEnumeratedLiteral */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_CommonEnumeratedLiteral, value),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonEnumeratedLiteral, flags), 0u, 127u,
  DDS_OP_RTS,

  /* CompleteBitmaskType */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CompleteBitmaskType, bitmask_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitmaskType, header), (3u << 16u) + 65504u /* CompleteEnumeratedHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_CompleteBitmaskType, flag_seq), sizeof (DDS_XTypes_CompleteBitflag), (4u << 16u) + 5u /* CompleteBitflag */,
  DDS_OP_RTS,

  /* CompleteBitflag */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitflag, common), (3u << 16u) + 7u /* CommonBitflag */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_CompleteBitflag, detail), (3u << 16u) + 65285u /* CompleteMemberDetail */,
  DDS_OP_RTS,

  /* CommonBitflag */
  DDS_OP_ADR | DDS_OP_TYPE_2BY, offsetof (DDS_XTypes_CommonBitflag, position),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_CommonBitflag, flags), 0u, 127u,
  DDS_OP_RTS,

  /* CompleteExtendedType */
  DDS_OP_PLC,
  DDS_OP_RTS,

  /* MinimalTypeObject */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY | DDS_OP_FLAG_DEF, offsetof (DDS_XTypes_MinimalTypeObject, _d), 11u, (48u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 45 /* MinimalAliasType */, 48, offsetof (DDS_XTypes_MinimalTypeObject, _u.alias_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 59 /* MinimalAnnotationType */, 80, offsetof (DDS_XTypes_MinimalTypeObject, _u.annotation_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 80 /* MinimalStructType */, 81, offsetof (DDS_XTypes_MinimalTypeObject, _u.struct_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 109 /* MinimalUnionType */, 82, offsetof (DDS_XTypes_MinimalTypeObject, _u.union_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 138 /* MinimalBitsetType */, 83, offsetof (DDS_XTypes_MinimalTypeObject, _u.bitset_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 157 /* MinimalSequenceType */, 96, offsetof (DDS_XTypes_MinimalTypeObject, _u.sequence_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 174 /* MinimalArrayType */, 97, offsetof (DDS_XTypes_MinimalTypeObject, _u.array_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 186 /* MinimalMapType */, 98, offsetof (DDS_XTypes_MinimalTypeObject, _u.map_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 196 /* MinimalEnumeratedType */, 64, offsetof (DDS_XTypes_MinimalTypeObject, _u.enumerated_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 217 /* MinimalBitmaskType */, 65, offsetof (DDS_XTypes_MinimalTypeObject, _u.bitmask_type), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 234 /* MinimalExtendedType */, 0, 0u, 0u,
  DDS_OP_RTS,

  /* MinimalAliasType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalAliasType, alias_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, 0u, (3u << 16u) + 7u /* MinimalAliasHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalAliasType, body), (3u << 16u) + 6u /* MinimalAliasBody */,
  DDS_OP_RTS,

  /* MinimalAliasHeader */
  DDS_OP_DLC,
  DDS_OP_RTS,

  /* MinimalAliasBody */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalAliasBody, common), (3u << 16u) + 65113u /* CommonAliasBody */,
  DDS_OP_RTS,

  /* MinimalAnnotationType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalAnnotationType, annotation_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, 0u, (3u << 16u) + 8u /* MinimalAnnotationHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_MinimalAnnotationType, member_seq), sizeof (DDS_XTypes_MinimalAnnotationParameter), (4u << 16u) + 7u /* MinimalAnnotationParameter */,
  DDS_OP_RTS,

  /* MinimalAnnotationHeader */
  DDS_OP_DLC,
  DDS_OP_RTS,

  /* MinimalAnnotationParameter */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalAnnotationParameter, common), (3u << 16u) + 65144u /* CommonAnnotationParameter */,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_MinimalAnnotationParameter, name_hash), 4u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalAnnotationParameter, default_value), (3u << 16u) + 65008u /* AnnotationParameterValue */,
  DDS_OP_RTS,

  /* MinimalStructType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalStructType, struct_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalStructType, header), (3u << 16u) + 8u /* MinimalStructHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_MinimalStructType, member_seq), sizeof (DDS_XTypes_MinimalStructMember), (4u << 16u) + 14u /* MinimalStructMember */,
  DDS_OP_RTS,

  /* MinimalStructHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalStructHeader, base_type), (3u << 16u) + 64807u /* TypeIdentifier */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, 0u, (3u << 16u) + 4u /* MinimalTypeDetail */,
  DDS_OP_RTS,

  /* MinimalTypeDetail */
  DDS_OP_RTS,

  /* MinimalStructMember */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalStructMember, common), (3u << 16u) + 65148u /* CommonStructMember */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalStructMember, detail), (3u << 16u) + 4u /* MinimalMemberDetail */,
  DDS_OP_RTS,

  /* MinimalMemberDetail */
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_MinimalMemberDetail, name_hash), 4u,
  DDS_OP_RTS,

  /* MinimalUnionType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalUnionType, union_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, 0u, (3u << 16u) + 11u /* MinimalUnionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalUnionType, discriminator), (3u << 16u) + 13u /* MinimalDiscriminatorMember */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_MinimalUnionType, member_seq), sizeof (DDS_XTypes_MinimalUnionMember), (4u << 16u) + 15u /* MinimalUnionMember */,
  DDS_OP_RTS,

  /* MinimalUnionHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, 0u, (3u << 16u) + 65507u /* MinimalTypeDetail */,
  DDS_OP_RTS,

  /* MinimalDiscriminatorMember */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalDiscriminatorMember, common), (3u << 16u) + 65171u /* CommonDiscriminatorMember */,
  DDS_OP_RTS,

  /* MinimalUnionMember */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalUnionMember, common), (3u << 16u) + 65182u /* CommonUnionMember */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalUnionMember, detail), (3u << 16u) + 65503u /* MinimalMemberDetail */,
  DDS_OP_RTS,

  /* MinimalBitsetType */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalBitsetType, bitset_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, 0u, (3u << 16u) + 8u /* MinimalBitsetHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_MinimalBitsetType, field_seq), sizeof (DDS_XTypes_MinimalBitfield), (4u << 16u) + 7u /* MinimalBitfield */,
  DDS_OP_RTS,

  /* MinimalBitsetHeader */
  DDS_OP_DLC,
  DDS_OP_RTS,

  /* MinimalBitfield */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalBitfield, common), (3u << 16u) + 65197u /* CommonBitfield */,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_MinimalBitfield, name_hash), 4u,
  DDS_OP_RTS,

  /* MinimalSequenceType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalSequenceType, collection_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalSequenceType, header), (3u << 16u) + 7u /* MinimalCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalSequenceType, element), (3u << 16u) + 9u /* MinimalCollectionElement */,
  DDS_OP_RTS,

  /* MinimalCollectionHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalCollectionHeader, common), (3u << 16u) + 65209u /* CommonCollectionHeader */,
  DDS_OP_RTS,

  /* MinimalCollectionElement */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalCollectionElement, common), (3u << 16u) + 65215u /* CommonCollectionElement */,
  DDS_OP_RTS,

  /* MinimalArrayType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalArrayType, collection_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalArrayType, header), (3u << 16u) + 7u /* MinimalArrayHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalArrayType, element), (3u << 16u) + 65524u /* MinimalCollectionElement */,
  DDS_OP_RTS,

  /* MinimalArrayHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalArrayHeader, common), (3u << 16u) + 65236u /* CommonArrayHeader */,
  DDS_OP_RTS,

  /* MinimalMapType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalMapType, collection_flag), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalMapType, header), (3u << 16u) + 65506u /* MinimalCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalMapType, key), (3u << 16u) + 65508u /* MinimalCollectionElement */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalMapType, element), (3u << 16u) + 65505u /* MinimalCollectionElement */,
  DDS_OP_RTS,

  /* MinimalEnumeratedType */
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalEnumeratedType, enum_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalEnumeratedType, header), (3u << 16u) + 8u /* MinimalEnumeratedHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_MinimalEnumeratedType, literal_seq), sizeof (DDS_XTypes_MinimalEnumeratedLiteral), (4u << 16u) + 10u /* MinimalEnumeratedLiteral */,
  DDS_OP_RTS,

  /* MinimalEnumeratedHeader */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalEnumeratedHeader, common), (3u << 16u) + 65242u /* CommonEnumeratedHeader */,
  DDS_OP_RTS,

  /* MinimalEnumeratedLiteral */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalEnumeratedLiteral, common), (3u << 16u) + 65248u /* CommonEnumeratedLiteral */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalEnumeratedLiteral, detail), (3u << 16u) + 65404u /* MinimalMemberDetail */,
  DDS_OP_RTS,

  /* MinimalBitmaskType */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_MinimalBitmaskType, bitmask_flags), 0u, 31u,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalBitmaskType, header), (3u << 16u) + 65518u /* MinimalEnumeratedHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_MinimalBitmaskType, flag_seq), sizeof (DDS_XTypes_MinimalBitflag), (4u << 16u) + 5u /* MinimalBitflag */,
  DDS_OP_RTS,

  /* MinimalBitflag */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalBitflag, common), (3u << 16u) + 65256u /* CommonBitflag */,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_MinimalBitflag, detail), (3u << 16u) + 65383u /* MinimalMemberDetail */,
  DDS_OP_RTS,

  /* MinimalExtendedType */
  DDS_OP_PLC,
  DDS_OP_RTS
};

const dds_topic_descriptor_t DDS_XTypes_TypeObject_desc =
{
  .m_size = sizeof (DDS_XTypes_TypeObject),
  .m_align = dds_alignof (DDS_XTypes_TypeObject),
  .m_flagset = DDS_TOPIC_CONTAINS_UNION,
  .m_nkeys = 0u,
  .m_typename = "DDS::XTypes::TypeObject",
  .m_keys = NULL,
  .m_nops = 404,
  .m_ops = DDS_XTypes_TypeObject_ops,
  .m_meta = ""
};

static const uint32_t DDS_XTypes_TypeInformation_ops [] =
{
  /* TypeInformation */
  DDS_OP_PLC,
  DDS_OP_PLM | 5, 4097u,
  DDS_OP_PLM | 7, 4098u,
  DDS_OP_RTS,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_TypeInformation, minimal), (3u << 16u) + 8u /* TypeIdentifierWithDependencies */,
  DDS_OP_RTS,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_TypeInformation, complete), (3u << 16u) + 4u /* TypeIdentifierWithDependencies */,
  DDS_OP_RTS,

  /* TypeIdentifierWithDependencies */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_TypeIdentifierWithDependencies, typeid_with_size), (3u << 16u) + 10u /* TypeIdentifierWithSize */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_TypeIdentifierWithDependencies, dependent_typeid_count),
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_STU, offsetof (DDS_XTypes_TypeIdentifierWithDependencies, dependent_typeids), sizeof (DDS_XTypes_TypeIdentifierWithSize), (4u << 16u) + 5u /* TypeIdentifierWithSize */,
  DDS_OP_RTS,

  /* TypeIdentifierWithSize */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_TypeIdentifierWithSize, type_id), (3u << 16u) + 6u /* TypeIdentifier */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_TypeIdentifierWithSize, typeobject_serialized_size),
  DDS_OP_RTS,

  /* TypeIdentifier */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeIdentifier, _d), 13u, (60u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 57 /* StringSTypeDefn */, 112, offsetof (DDS_XTypes_TypeIdentifier, _u.string_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 53 /* StringSTypeDefn */, 114, offsetof (DDS_XTypes_TypeIdentifier, _u.string_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 52 /* StringLTypeDefn */, 113, offsetof (DDS_XTypes_TypeIdentifier, _u.string_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 48 /* StringLTypeDefn */, 115, offsetof (DDS_XTypes_TypeIdentifier, _u.string_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 47 /* PlainSequenceSElemDefn */, 128, offsetof (DDS_XTypes_TypeIdentifier, _u.seq_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 60 /* PlainSequenceLElemDefn */, 129, offsetof (DDS_XTypes_TypeIdentifier, _u.seq_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 66 /* PlainArraySElemDefn */, 144, offsetof (DDS_XTypes_TypeIdentifier, _u.array_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 72 /* PlainArrayLElemDefn */, 145, offsetof (DDS_XTypes_TypeIdentifier, _u.array_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 78 /* PlainMapSTypeDefn */, 160, offsetof (DDS_XTypes_TypeIdentifier, _u.map_sdefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 92 /* PlainMapLTypeDefn */, 161, offsetof (DDS_XTypes_TypeIdentifier, _u.map_ldefn), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_STU | 106 /* StronglyConnectedComponentId */, 176, offsetof (DDS_XTypes_TypeIdentifier, _u.sc_component_id), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 8, 242, offsetof (DDS_XTypes_TypeIdentifier, _u.equivalence_hash), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 4, 241, offsetof (DDS_XTypes_TypeIdentifier, _u.equivalence_hash), 0u,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, 0u, 14u,
  DDS_OP_RTS,
  DDS_OP_RTS,

  /* StringSTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_StringSTypeDefn, bound),
  DDS_OP_RTS,

  /* StringLTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_StringLTypeDefn, bound),
  DDS_OP_RTS,

  /* PlainSequenceSElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceSElemDefn, header), (3u << 16u) + 10u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainSequenceSElemDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceSElemDefn, element_identifier), (4u << 16u) + 65464u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainCollectionHeader */
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainCollectionHeader, equiv_kind),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainCollectionHeader, element_flags), 0u, 127u,
  DDS_OP_RTS,

  /* PlainSequenceLElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceLElemDefn, header), (3u << 16u) + 65529u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_PlainSequenceLElemDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainSequenceLElemDefn, element_identifier), (4u << 16u) + 65447u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainArraySElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArraySElemDefn, header), (3u << 16u) + 65519u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_PlainArraySElemDefn, array_bound_seq),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArraySElemDefn, element_identifier), (4u << 16u) + 65437u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainArrayLElemDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArrayLElemDefn, header), (3u << 16u) + 65509u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_SEQ | DDS_OP_SUBTYPE_4BY, offsetof (DDS_XTypes_PlainArrayLElemDefn, array_bound_seq),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainArrayLElemDefn, element_identifier), (4u << 16u) + 65427u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainMapSTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, header), (3u << 16u) + 65499u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_1BY, offsetof (DDS_XTypes_PlainMapSTypeDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, element_identifier), (4u << 16u) + 65417u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainMapSTypeDefn, key_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapSTypeDefn, key_identifier), (4u << 16u) + 65409u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* PlainMapLTypeDefn */
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, header), (3u << 16u) + 65481u /* PlainCollectionHeader */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY, offsetof (DDS_XTypes_PlainMapLTypeDefn, bound),
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, element_identifier), (4u << 16u) + 65399u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_ADR | DDS_OP_TYPE_BMK | (1 << DDS_OP_FLAG_SZ_SHIFT), offsetof (DDS_XTypes_PlainMapLTypeDefn, key_flags), 0u, 127u,
  DDS_OP_ADR | DDS_OP_FLAG_EXT | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_PlainMapLTypeDefn, key_identifier), (4u << 16u) + 65391u /* TypeIdentifier */, sizeof (DDS_XTypes_TypeIdentifier),
  DDS_OP_RTS,

  /* StronglyConnectedComponentId */
  DDS_OP_DLC,
  DDS_OP_ADR | DDS_OP_TYPE_EXT, offsetof (DDS_XTypes_StronglyConnectedComponentId, sc_component_id), (3u << 16u) + 8u /* TypeObjectHashId */,
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_StronglyConnectedComponentId, scc_length),
  DDS_OP_ADR | DDS_OP_TYPE_4BY | DDS_OP_FLAG_SGN, offsetof (DDS_XTypes_StronglyConnectedComponentId, scc_index),
  DDS_OP_RTS,

  /* TypeObjectHashId */
  DDS_OP_ADR | DDS_OP_FLAG_MU | DDS_OP_TYPE_UNI | DDS_OP_SUBTYPE_1BY, offsetof (DDS_XTypes_TypeObjectHashId, _d), 2u, (16u << 16u) + 4u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 8, 242, offsetof (DDS_XTypes_TypeObjectHashId, _u.hash), 0u,
  DDS_OP_JEQ4 | DDS_OP_TYPE_ARR | 4, 241, offsetof (DDS_XTypes_TypeObjectHashId, _u.hash), 0u,
  DDS_OP_ADR | DDS_OP_TYPE_ARR | DDS_OP_SUBTYPE_1BY, 0u, 14u,
  DDS_OP_RTS,
  DDS_OP_RTS
};

const dds_topic_descriptor_t DDS_XTypes_TypeInformation_desc =
{
  .m_size = sizeof (DDS_XTypes_TypeInformation),
  .m_align = dds_alignof (DDS_XTypes_TypeInformation),
  .m_flagset = DDS_TOPIC_CONTAINS_UNION,
  .m_nkeys = 0u,
  .m_typename = "DDS::XTypes::TypeInformation",
  .m_keys = NULL,
  .m_nops = 78,
  .m_ops = DDS_XTypes_TypeInformation_ops,
  .m_meta = ""
};
