/*
 * Copyright(c) 2006 to 2021 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
 */
#ifndef DDS_VERSION_H
#define DDS_VERSION_H

#define DDS_VERSION "0.10.4"
#define DDS_VERSION_MAJOR 0
#define DDS_VERSION_MINOR 10
#define DDS_VERSION_PATCH 4
#define DDS_VERSION_TWEAK 
#define DDS_PROJECT_NAME "CycloneDDS"

#define DDS_HOST_NAME "Linux"
#define DDS_TARGET_NAME "Linux"

#endif /* DDS_VERSION_H */
